lnd_config_template = \
    "[Application Options]\n" \
    "alias={name}\n" \
    "listen=localhost:{lnd_port}\n" \
    "restlisten=localhost:{rest_port}\n" \
    "rpclisten=localhost:{rpc_port}\n" \
    "debuglevel=debug\n" \
    "unsafe-disconnect=1\n" \
    "[Bitcoin]\n" \
    "bitcoin.active=1\n" \
    "bitcoin.regtest=1\n" \
    "bitcoin.node=bitcoind\n" \
    "bitcoin.basefee={base_fee_msat}\n" \
    "bitcoin.feerate={fee_rate}\n" \
    "[Bitcoind]\n" \
    "bitcoind.rpchost=localhost\n" \
    "bitcoind.rpcuser=lnd\n" \
    "bitcoind.rpcpass=123456\n" \
    "bitcoind.zmqpubrawblock=tcp://127.0.0.1:28332\n" \
    "bitcoind.zmqpubrawtx=tcp://127.0.0.1:28333\n" \
    "[protocol]\n" \
    "protocol.wumbo-channels=true"

bitcoind_config_template = \
    "txindex=1\n" \
    "zmqpubrawblock=tcp://127.0.0.1:28332\n" \
    "zmqpubrawtx=tcp://127.0.0.1:28333\n" \
    "regtest=1\n" \
    "rpcuser=lnd\n" \
    "rpcpassword=123456\n" \
    "fallbackfee=0.00001000\n"



import os
import codecs
import time
import datetime
import statistics
from typing import List, TYPE_CHECKING

from collections import OrderedDict

import grpc
from grpc._channel import _Rendezvous
from google.protobuf.json_format import MessageToDict

import lndmanage.grpc_compiled.rpc_pb2 as lnd
import lndmanage.grpc_compiled.rpc_pb2_grpc as lndrpc

import lndmanage.grpc_compiled.router_pb2 as lndrouter
import lndmanage.grpc_compiled.router_pb2_grpc as lndrouterrpc

from lndmanage.lib.network import Network
from lndmanage.lib.exceptions import PaymentTimeOut, NoRoute, OurNodeFailure
from lndmanage.lib import exceptions
from lndmanage.lib.utilities import convert_dictionary_number_strings_to_ints
from lndmanage.lib.ln_utilities import (
    extract_short_channel_id_from_string,
    convert_short_channel_id_to_channel_id,
    convert_channel_id_to_short_channel_id,
    channel_unbalancedness_and_commit_fee
)
from lndmanage import settings

if TYPE_CHECKING:
    from lndmanage.lib.routing import Route

import logging
logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())

NUM_MAX_FORWARDING_EVENTS = 100000


class Node(object):
    """Bare node object with attributes."""
    def __init__(self):
        logger.info("Initializing node interface.")
        self.alias = ''
        self.pub_key = ''
        self.total_capacity = 0
        self.total_local_balance = 0
        self.total_remote_balance = 0
        self.total_channels = 0
        self.num_active_channels = 0
        self.num_peers = 0
        self.total_satoshis_received = 0
        self.total_satoshis_sent = 0
        self.total_private_channels = 0
        self.total_active_channels = 0
        self.blockheight = 0
        self.channels = []


class LndNode(Node):
    """
    Implements an interface to an lnd node.
    """
    def __init__(self, config_file=None, lnd_home=None, lnd_host=None,
                 regtest=False):
        """
        :param config_file: path to the config file
        :type config_file: str
        :param lnd_home: path to lnd home folder
        :type lnd_home: str
        :param lnd_host: lnd host of format "127.0.0.1:9735"
        :type lnd_host: str
        :param regtest: if the node is representing a regtest node
        :type regtest: bool
        """
        super().__init__()
        self.config_file = config_file
        self.lnd_home = lnd_home
        self.lnd_host = lnd_host
        self.regtest = regtest

        self._rpc = None
        self._routerrpc = None
        self.connect_rpcs()

        self.network = Network(self)
        self.update_blockheight()
        self.set_info()

    def connect_rpcs(self):
        """
        Establishes a connection to lnd using the hostname, tls certificate,
        and admin macaroon defined in settings.
        """
        macaroons = True
        os.environ['GRPC_SSL_CIPHER_SUITES'] = \
            'ECDHE-RSA-AES128-GCM-SHA256:' \
            'ECDHE-RSA-AES128-SHA256:' \
            'ECDHE-RSA-AES256-SHA384:' \
            'ECDHE-RSA-AES256-GCM-SHA384:' \
            'ECDHE-ECDSA-AES128-GCM-SHA256:' \
            'ECDHE-ECDSA-AES128-SHA256:' \
            'ECDHE-ECDSA-AES256-SHA384:' \
            'ECDHE-ECDSA-AES256-GCM-SHA384'

        # if no lnd_home is given, then use the paths from the config,
        # else override them with default file paths in lnd_home
        if self.lnd_home is not None:
            cert_file = os.path.join(self.lnd_home, 'tls.cert')
            bitcoin_network = 'regtest' if self.regtest else 'mainnet'
            macaroon_file = os.path.join(
                self.lnd_home, 'data/chain/bitcoin/',
                bitcoin_network, 'admin.macaroon')
            if self.lnd_host is None:
                raise ValueError(
                    'if lnd_home is given, lnd_host must be given also')
            lnd_host = self.lnd_host
        else:
            config = settings.read_config(self.config_file)
            cert_file = os.path.expanduser(config['network']['tls_cert_file'])
            macaroon_file = \
                os.path.expanduser(config['network']['admin_macaroon_file'])
            lnd_host = config['network']['lnd_grpc_host']

        cert = None
        try:
            with open(cert_file, 'rb') as f:
                cert = f.read()
        except FileNotFoundError:
            logger.error("tls.cert not found, please configure %s.",
                         self.config_file)
            exit(1)

        if macaroons:
            try:
                with open(macaroon_file, 'rb') as f:
                    macaroon_bytes = f.read()
                    macaroon = codecs.encode(macaroon_bytes, 'hex')
            except FileNotFoundError:
                logger.error("admin.macaroon not found, please configure %s.",
                             self.config_file)
                exit(1)

            def metadata_callback(context, callback):
                callback([('macaroon', macaroon)], None)

            cert_creds = grpc.ssl_channel_credentials(cert)
            auth_creds = grpc.metadata_call_credentials(metadata_callback)
            creds = grpc.composite_channel_credentials(cert_creds, auth_creds)

        else:
            creds = grpc.ssl_channel_credentials(cert)

        # necessary to circumvent standard size limitation
        channel = grpc.secure_channel(lnd_host, creds, options=[
            ('grpc.max_receive_message_length', 50 * 1024 * 1024)
        ])

        # establish connections to rpc servers
        self._rpc = lndrpc.LightningStub(channel)
        self._routerrpc = lndrouterrpc.RouterStub(channel)

    def update_blockheight(self):
        info = self._rpc.GetInfo(lnd.GetInfoRequest())
        self.blockheight = int(info.block_height)

    def get_channel_info(self, channel_id):
        channel = self._rpc.GetChanInfo(
            lnd.ChanInfoRequest(chan_id=channel_id))
        channel_dict = MessageToDict(
            channel, including_default_value_fields=True)
        channel_dict = convert_dictionary_number_strings_to_ints(channel_dict)
        return channel_dict

    @staticmethod
    def lnd_hops(hops) -> List[lnd.Hop]:
        return [lnd.Hop(**hop) for hop in hops]

    def _to_lnd_route(self, route: 'Route') -> lnd.Route:
        """
        Converts a cleartext route to an lnd route.
        """
        hops = self.lnd_hops(route.hops)
        lnd_route = lnd.Route(
            total_time_lock=route.total_time_lock,
            total_fees=route.total_fee_msat // 1000,
            total_amt=route.total_amt_msat // 1000,
            hops=hops,
            total_fees_msat=route.total_fee_msat,
            total_amt_msat=route.total_amt_msat
        )
        return lnd_route

    def get_invoice(self, amt_msat: int, memo: str) -> lnd.Invoice:
        """
        Creates a new invoice with amt_msat and memo.

        :param amt_msat: int
        :param memo: str
        :return: Hash of invoice preimage.
        """
        invoice = self._rpc.AddInvoice(lnd.Invoice(
            value=amt_msat // 1000, memo=memo))
        return invoice

    def get_rebalance_invoice(self, memo) -> lnd.Invoice:
        """
        Creates a zero amount invoice and gives back it's hash.

        :param memo: Comment for the invoice.
        :return: Hash of the invoice preimage.
        """
        invoice = self._rpc.AddInvoice(lnd.Invoice(value=0, memo=memo))
        return invoice

    def send_to_route(self, route: 'Route', payment_hash: bytes,
                      payment_address: bytes):
        """
        Takes bare route (list) and tries to send along it,
        trying to fulfill the invoice labeled by the given hash.

        :param route: (list) of :class:`lib.routes.Route`
        :param payment_hash: invoice identifier
        :return:
        """
        lnd_route = self._to_lnd_route(route)
        # set payment address for last hop
        lnd_route.hops[-1].tlv_payload = True
        lnd_route.hops[-1].mpp_record.payment_addr = payment_address
        lnd_route.hops[-1].mpp_record.total_amt_msat = lnd_route.hops[-1].amt_to_forward_msat
        # set payment hash
        request = lndrouter.SendToRouteRequest(
            route=lnd_route,
            payment_hash=payment_hash,
        )
        try:
            # timeout after 5 minutes
            payment = self._routerrpc.SendToRouteV2(request, timeout=5 * 60)
        except _Rendezvous:
            raise PaymentTimeOut
        if payment.HasField('failure'):
            failure = payment.failure  # type: lnd.Failure.FailureCode
            logger.debug(f"Routing failure: {failure}")
            if failure.failure_source_index == 0:
                raise OurNodeFailure("Not enough funds?")
            if failure.code == 15:
                raise exceptions.TemporaryChannelFailure(payment)
            elif failure.code == 14:
                raise exceptions.ChannelDisabled(payment)
            elif failure.code == 18:
                raise exceptions.UnknownNextPeer(payment)
            elif failure.code == 12:
                raise exceptions.FeeInsufficient(payment)
            else:
                logger.info(failure)
                raise Exception(f"Unknown error: code {failure.code}")

        return payment

    def get_raw_network_graph(self):
        try:
            graph = self._rpc.DescribeGraph(lnd.ChannelGraphRequest())
            return graph
        except _Rendezvous:
            logger.error(
                "Problem connecting to lnd. "
                "Either %s is not configured correctly or lnd is not running.",
                self.config_file)
            exit(1)

    def get_raw_info(self):
        """
        Returns specific information about this node.

        :return: node information
        """
        return self._rpc.GetInfo(lnd.GetInfoRequest())

    def set_info(self):
        """
        Fetches information about this node and computes total capacity,
        local and remote total balance, how many satoshis were sent and
        received, and some networking peer stats.
        """

        raw_info = self.get_raw_info()
        self.pub_key = raw_info.identity_pubkey
        self.alias = raw_info.alias
        self.num_active_channels = raw_info.num_active_channels
        self.num_peers = raw_info.num_peers

        # TODO: remove the following code and implement an advanced status
        all_channels = self.get_open_channels(
            active_only=False, public_only=False)
        self.total_channels = len(all_channels)

        for k, c in all_channels.items():
            self.total_capacity += c['capacity']
            self.total_local_balance += c['local_balance']
            self.total_remote_balance += c['remote_balance']
            self.total_satoshis_received += c['total_satoshis_received']
            self.total_satoshis_sent += c['total_satoshis_sent']
            if c['active']:
                self.total_active_channels += 1
            if c['private']:
                self.total_private_channels += 1

    def get_open_channels(self, active_only=False, public_only=False):
        """
        Fetches information (fee settings of the counterparty, channel
        capacity, balancedness) about this node's open channels and saves
        it into the channels dict attribute.

        :param active_only: only take active channels into
                            account (off by default)
        :type active_only: bool
        :param public_only: only take public channels into
                            account (off by default)
        :type public_only: bool

        :return: dict of channels sorted by remote pubkey
        :rtype: OrderedDict

        """
        raw_channels = self._rpc.ListChannels(lnd.ListChannelsRequest(
            active_only=active_only, public_only=public_only))
        try:
            channels_data = raw_channels.ListFields()[0][1]
        except IndexError:
            # If there are no channels, return.
            return OrderedDict({})

        channels = OrderedDict()

        for c in channels_data:
            # calculate age from blockheight
            blockheight, _, _ = convert_channel_id_to_short_channel_id(
                c.chan_id)
            age_days = (self.blockheight - blockheight) * 10 / (60 * 24)
            try:
                sent_received_per_week = int(
                    (c.total_satoshis_sent +
                     c.total_satoshis_received) / (age_days / 7))
            except ZeroDivisionError:
                # age could be zero right after channel becomes pending
                sent_received_per_week = 0

            # determine policy
            try:
                edge_info = self.network.edges[c.chan_id]
                # interested in node2
                if edge_info['node1_pub'] == self.pub_key:
                    policy_peer = edge_info['node2_policy']
                    policy_local = edge_info['node1_policy']
                else:  # interested in node1
                    policy_peer = edge_info['node1_policy']
                    policy_local = edge_info['node2_policy']
            except KeyError:
                # if channel is unknown in describegraph
                # we need to set the fees to some error value
                policy_peer = {'fee_base_msat': float(-999),
                               'fee_rate_milli_msat': float(999)}
                policy_local = {'fee_base_msat': float(-999),
                                'fee_rate_milli_msat': float(999)}

            # calculate last update (days ago)
            def convert_to_days_ago(timestamp):
                return (time.time() - timestamp) / (60 * 60 * 24)
            try:
                last_update = convert_to_days_ago(
                    self.network.edges[c.chan_id]['last_update'])
                last_update_local = convert_to_days_ago(
                    policy_local['last_update'])
                last_update_peer = convert_to_days_ago(
                    policy_peer['last_update'])
            except (TypeError, KeyError):
                last_update = float('nan')
                last_update_peer = float('nan')
                last_update_local = float('nan')

            # define unbalancedness |ub| large means very unbalanced
            channel_unbalancedness, our_commit_fee = \
                channel_unbalancedness_and_commit_fee(
                    c.local_balance, c.capacity, c.commit_fee, c.initiator)
            try:
                uptime_lifetime_ratio = c.uptime / c.lifetime
            except ZeroDivisionError:
                uptime_lifetime_ratio = 0

            channels[c.chan_id] = {
                'active': c.active,
                'age': age_days,
                'alias': self.network.node_alias(c.remote_pubkey),
                'amt_to_balanced': int(
                    channel_unbalancedness * c.capacity / 2 - our_commit_fee),
                'capacity': c.capacity,
                'chan_id': c.chan_id,
                'channel_point': c.channel_point,
                'commit_fee': c.commit_fee,
                'fee_per_kw': c.fee_per_kw,
                'peer_base_fee': policy_peer['fee_base_msat'],
                'peer_fee_rate': policy_peer['fee_rate_milli_msat'],
                'local_base_fee': policy_local['fee_base_msat'],
                'local_fee_rate': policy_local['fee_rate_milli_msat'],
                'initiator': c.initiator,
                'last_update': last_update,
                'last_update_local': last_update_local,
                'last_update_peer': last_update_peer,
                'local_balance': c.local_balance,
                'num_updates': c.num_updates,
                'private': c.private,
                'remote_balance': c.remote_balance,
                'remote_pubkey': c.remote_pubkey,
                'sent_received_per_week': sent_received_per_week,
                'total_satoshis_sent': c.total_satoshis_sent,
                'total_satoshis_received': c.total_satoshis_received,
                'unbalancedness': channel_unbalancedness,
                'uptime': c.uptime,
                'lifetime': c.lifetime,
                'uptime_lifetime_ratio': uptime_lifetime_ratio,
            }
        sorted_dict = OrderedDict(
            sorted(channels.items(), key=lambda x: x[1]['alias']))
        return sorted_dict

    def get_inactive_channels(self):
        """
        Returns all inactive channels.
        :return: dict of channels
        """
        channels = self.get_open_channels(public_only=False, active_only=False)
        return {k: c for k, c in channels.items() if not c['active']}

    def get_all_channels(self):
        """
        Returns all active and inactive channels.

        :return: dict of channels
        """
        channels = self.get_open_channels(public_only=False, active_only=False)
        return channels

    def get_unbalanced_channels(self, unbalancedness_greater_than=0.0):
        """
        Gets all channels which have an absolute unbalancedness
        (-1...1, -1 for outbound unbalanced, 1 for inbound unbalanced)
        larger than unbalancedness_greater_than.

        :param unbalancedness_greater_than: unbalancedness interval, default returns all channels
        :return: all channels which are more unbalanced than the specified interval
        """
        self.public_active_channels = \
            self.get_open_channels(public_only=True, active_only=True)
        unbalanced_channels = {
            k: c for k, c in self.public_active_channels.items()
            if abs(c['unbalancedness']) >= unbalancedness_greater_than
        }
        return unbalanced_channels

    @staticmethod
    def timestamp_from_now(offset_days=0):
        """
        Determines the Unix timestamp from offset_days ago.

        :param offset_days: int
        :return: int, Unix timestamp
        """
        now = datetime.datetime.now()
        then = now - datetime.timedelta(days=offset_days)
        then = time.mktime(then.timetuple())
        return int(then)

    def get_forwarding_events(self, offset_days=300):
        """
        Fetches all forwarding events between now and offset_days ago.

        :param offset_days: int
        :return: lnd fowarding events
        """
        now = self.timestamp_from_now()
        then = self.timestamp_from_now(offset_days)

        forwardings = self._rpc.ForwardingHistory(lnd.ForwardingHistoryRequest(
            start_time=then,
            end_time=now,
            num_max_events=NUM_MAX_FORWARDING_EVENTS))

        events = [{
            'timestamp': f.timestamp,
            'chan_id_in': f.chan_id_in,
            'chan_id_out': f.chan_id_out,
            'amt_in': f.amt_in,
            'amt_in_msat': f.amt_in_msat,
            'amt_out': f.amt_out,
            'amt_out_msat': f.amt_out_msat,
            'fee_msat': f.fee_msat,
            'effective_fee': f.fee_msat / f.amt_in_msat
        } for f in forwardings.forwarding_events]

        return events

    def get_closed_channels(self):
        """
        Fetches all closed channels.

        :return: dict, channel list
        """
        request = lnd.ClosedChannelsRequest()
        closed_channels = self._rpc.ClosedChannels(request)
        closed_channels_dict = {}
        for c in closed_channels.channels:
            closed_channels_dict[c.chan_id] = {
                'channel_point': c.channel_point,
                'chain_hash': c.chain_hash,
                'closing_tx_hash': c.closing_tx_hash,
                'remote_pubkey': c.remote_pubkey,
                'capacity': c.capacity,
                'close_height': c.close_height,
                'settled_balance': c.settled_balance,
                'time_locked_balance': c.time_locked_balance,
                'close_type': c.close_type,
            }
        return closed_channels_dict

    @staticmethod
    def handle_payment_error(payment_error):
        """
        Handles payment errors and determines the failed channel.

        :param payment_error:
        :return: int, channel_id of the failed channel.
        """
        if "TemporaryChannelFailure" in payment_error:
            logger.error("   Encountered temporary channel failure.")
            short_channel_groups = extract_short_channel_id_from_string(payment_error)
            channel_id = convert_short_channel_id_to_channel_id(*short_channel_groups)
            return channel_id

    def queryroute_external(self, source_pubkey, target_pubkey, amt_msat,
                            ignored_nodes=(), ignored_channels={},
                            use_mc=False):
        """
        Queries the lnd node for a route.

        Channels and nodes can be ignored if they failed before.

        :param source_pubkey: source node public key
        :type source_pubkey: str
        :param target_pubkey: target node public key
        :type target_pubkey: str
        :param amt_msat: amount to send in msat
        :type amt_msat: int
        :param ignored_nodes: ignored node pubilc keys for the route
        :type ignored_nodes: list[str]
        :param ignored_channels: ignored channel directions for the route
        :type ignored_channels: dict
        :param use_mc: true if mission control should be used to blacklist
                       channels
        :type use_mc: bool
        :return: route expressed in terms of short channel ids
        :rtype: list[int]
        """
        amt_sat = amt_msat // 1000

        # put safety margin when using mc based routing
        # reason is that routes will not be diverse when sending with a larger
        # amount later on due to fees
        # the fees for the route are somewhat accounted for by the margin
        if use_mc:
            amt_sat = int(amt_sat * 1.02)

        # have a safety max fee in sat
        max_fee = 10000

        # convert ignored nodes to api format
        if ignored_nodes:
            ignored_nodes_api = [bytes.fromhex(n) for n in ignored_nodes]
        else:
            ignored_nodes_api = []

        # convert ignored channels to api format
        if ignored_channels:
            ignored_channels_api = []
            for c, cv in ignored_channels.items():
                direction_reverse = cv['source'] > cv['target']
                ignored_channels_api.append(
                    lnd.EdgeLocator(channel_id=c,
                                    direction_reverse=direction_reverse))
        else:
            ignored_channels_api = []

        logger.debug(f"Ignored for queryroutes: channels: "
                     f"{ignored_channels_api}, nodes: {ignored_nodes_api}")

        request = lnd.QueryRoutesRequest(
            pub_key=target_pubkey,
            amt=amt_sat,
            final_cltv_delta=0,
            fee_limit=lnd.FeeLimit(fixed=max_fee),
            ignored_nodes=ignored_nodes_api,
            ignored_edges=ignored_channels_api,
            source_pub_key=source_pubkey,
            use_mission_control=use_mc,
        )
        try:
            response = self._rpc.QueryRoutes(request)
        except Exception as e:
            if "unable to find a path" in e.details():
                raise NoRoute
            else:
                raise e

        # We give back only one route, as multiple routes will be deprecated
        channel_route = [h.chan_id for h in response.routes[0].hops]

        return channel_route

    def get_node_info(self, pub_key):
        """
        Retrieves information on a node with a specific pub key.

        :param pub_key: node public key
        :type pub_key: str
        :return: node information including all channels
        :rtype: dict
        """
        request = lnd.NodeInfoRequest(pub_key=pub_key, include_channels=True)
        try:
            response = self._rpc.GetNodeInfo(request)
        except _Rendezvous as e:
            if e.details() == "unable to find node":
                logger.info(
                    "LND node has no information about node with pub key %s.",
                    pub_key)
            raise KeyError

        node_info = {
            'alias': response.node.alias,
            'color': response.node.color,
            'channels': response.channels,
            'last_update': response.node.last_update,
            'pub_key': pub_key,
            'num_channels': int(response.num_channels),
            'total_capacity': int(response.total_capacity),  # sat
        }

        addresses = [address.addr for address in response.node.addresses]
        node_info['addresses'] = addresses

        return node_info

    def print_status(self):
        logger.info("-------- Node status --------")
        if self.total_capacity == 0:
            balancedness_local = 0
            balancedness_remote = 0
        else:
            balancedness_local = self.total_local_balance / self.total_capacity
            balancedness_remote = self.total_remote_balance / self.total_capacity
        logger.info(f"alias: {self.alias}")
        logger.info(f"pub key: {self.pub_key}")
        logger.info(f"blockheight: {self.blockheight}")
        logger.info(f"peers: {self.num_peers}")
        logger.info(f"channels: {self.total_channels}")
        logger.info(f"active channels: {self.total_active_channels}")
        logger.info(f"private channels: {self.total_private_channels}")
        logger.info(f"capacity: {self.total_capacity}")
        logger.info(f"balancedness: l:{balancedness_local:.2%} r:{balancedness_remote:.2%}")
        logger.info(f"total satoshis received (current channels): {self.total_satoshis_received}")
        logger.info(f"total satoshis sent (current channels): {self.total_satoshis_sent}")


if __name__ == '__main__':
    node = LndNode()
    print(node.get_closed_channels().keys())



"""
Module for implementing a bitcoin/lightning regtest network.
"""
from collections import defaultdict
import importlib
import importlib.util
import logging
import os
import pickle
from pprint import pprint, pformat
import tempfile
import time
from typing import Mapping, List, Optional, Union

from lnregtest.lib.common import (
    logger_config, WAIT_AFTER_MINING_THREE, WAIT_AFTER_ALL_LND_STARTED,
    WAIT_AFTER_FILLING_WALLETS, WAIT_BEFORE_CLEANUP
)
from lnregtest.lib.network_components import (Bitcoind, LND, LightningDaemon,
                                              Electrum, ElectrumX)
from lnregtest.lib.utils import format_dict, convert_short_channel_id_to_channel_id, bfh
from lnregtest.lib.graph_testing import graph_test

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())

AVAILABLE_DAEMONS = {
    'lnd': LND,
    'electrum': Electrum,
}


class Network(object):
    """
    Wires all the network components together and controls the logic.
    """
    def __init__(self, binary_folder=None,
                 network_definition_location='star_ring',
                 nodedata_folder='', node_limit='C', from_scratch=True):
        """
        :param binary_folder:str:
            absolute path to where node/cli binaries reside, if not given,
            binaries will be taken from $PATH
        :param network_definition_location: str:
            specifies python module in network_templates which defines the
            network
            alternatively, an absolute path to a python
        :param nodedata_folder: str:
            absolute path to nodedata folder, where node runtime data is saved
            if no value is given, a temporary folder is created
        :param node_limit: char:
            even if more nodes are specified in the network definition,
            the number of nodes used can be limited by specifying an upper
            limit indicated by a character, e.g. 'C' would create nodes A, B, C
        :param from_scratch: bool:
            specifies if the network should be generated without restoring from
            a previous state
        """
        # determine the network_definition
        if os.path.isabs(network_definition_location):
            # if absolute path, import from absolute path
            spec = importlib.util.spec_from_file_location(
                "network_definition",
                network_definition_location)

            network_definition_module = importlib.util.module_from_spec(spec)
            spec.loader.exec_module(network_definition_module)
            # extract the file name from the path and use as suffix for
            # runtime data folder
            file_without_extension = os.path.splitext(
                network_definition_location)[0]
            network_definition_suffix = os.path.basename(
                file_without_extension)
        else:
            # if network_definition is not an absolute path
            network_definition_suffix = network_definition_location
            network_definition_module = importlib.import_module(
                ".network_definitions." + network_definition_location,
                package='lnregtest')

        # define where runtime data will be saved to
        if not nodedata_folder:
            if not from_scratch:
                raise ValueError(
                    'If no nodedata_folder is given, '
                    'restarting is not possible.')
            # create temporary directory
            self.temp_dir = tempfile.TemporaryDirectory(
                prefix='lnregtest-', suffix='-' + network_definition_suffix)
            self.nodedata_folder = self.temp_dir.name
        else:
            self.nodedata_folder = os.path.join(
                nodedata_folder, network_definition_suffix)
        logger.info('Runtime data resides in: %s', self.nodedata_folder)

        # check sanity of network definition
        graph_test(network_definition_module.nodes)

        # initialize bitcoind
        self.from_scratch = from_scratch
        self.node_limit = node_limit
        self.bitcoind = Bitcoind(self.nodedata_folder, binary_folder)

        # get network definition
        self.network_definition = self.get_reduced_network_definition(
            network_definition_module)

        # check if there's an electrum node
        daemons = [n.get('daemon') for n in self.network_definition.values()]

        # initialize electrumx
        if 'electrum' in daemons:
            logger.info(f"NET: Found electrum, need electrumx.")
            self.electrumx = ElectrumX(self.nodedata_folder)
        else:
            self.electrumx = None

        # define empty node and channel mappings
        self.channel_mapping = defaultdict(dict)
        self.channel_mapping_inverse = {}
        self.node_mapping = {}
        self.node_mapping_inverse = {}

        # initialize lightning nodes
        self.running = False

        self.ln_nodes: Mapping[str, Union[Electrum, LND]] = {}
        for node_name, node_properties in self.network_definition.items():
            if node_properties['daemon'] == 'electrum':
                LNDaemon = Electrum
            else:
                LNDaemon = LND

            self.ln_nodes[node_name] = LNDaemon(
                name=node_name,
                node_properties=node_properties,
                nodes_folder=self.nodedata_folder,
                binary_folder=binary_folder,
            )

        self.master_node = self.ln_nodes['A']

        # define paths for node and channel mappings
        self.node_mapping_path = os.path.join(
            self.nodedata_folder, 'node_mapping.pickle')
        self.channel_mapping_path = os.path.join(
            self.nodedata_folder, 'channel_mapping.pickle')

        # read in channel and node mappings
        if not from_scratch:
            self.read_mappings()

    def run_nocleanup(self):
        """
        Runs the daemons without cleaning up in the end. Use in conjunction
        with self.cleanup() (try-finally).
        """
        logger.info("Running regtest network (from scratch: %s).",
                    self.from_scratch)
        # start bitcoind
        self.bitcoind.start(self.from_scratch)

        if self.electrumx:
            self.electrumx.start()

        if self.from_scratch:
            # fill the bitcoind wallet
            self.bitcoind.fill_addresses(10)
            self.bitcoind.mine_blocks(100)

        self.nodes_start()

        # delay to get all started up
        time.sleep(WAIT_AFTER_ALL_LND_STARTED)
        self.determine_node_mapping()

        # open channels and mine blocks for confirmation
        if self.from_scratch:
            self.nodes_fill_wallets()
            time.sleep(WAIT_AFTER_FILLING_WALLETS)
            self.nodes_connect_open_channels()
            # finalize last channel
            self.bitcoind.mine_blocks(3)
            time.sleep(3)

        # show state of lightning nodes
        self.nodes_print_info()

        logger.info(f"NET: node mapping\n{pformat(self.node_mapping)}")
        # determine the mapping from channels to channel numbers
        self.determine_channel_mapping()

        if self.from_scratch:
            self.save_mappings()
        else:
            self.read_mappings()

        # set fees on a per node basis
        # self.nodes_set_fees()

        self.master_node_graph_view()

        self.running = True

        logger.info("\nLocal lightning network is running. Have fun!")
        self.print_cli_commands()

    def run_once(self):
        """
        Run the network and terminate after.
        """
        try:
            self.run_nocleanup()
        finally:
            self.cleanup()

    def run_continuously(self):
        """
        Run the network continuously.
        """
        try:
            self.run_nocleanup()
            while True:
                time.sleep(2)
        except KeyboardInterrupt:
            logger.info("Will shut down lnregtest.")
        finally:
            self.cleanup()

    def run_from_background(self):
        self.read_mappings()

    def cleanup(self):
        time.sleep(WAIT_BEFORE_CLEANUP)
        self.stop_components()
        try:
            self.temp_dir.cleanup()
        except AttributeError:
            logger.debug(
                "Will keep network data, as nodedata_folder is given.")

    def stop_components(self):
        self.nodes_stop()
        if self.electrumx:
            self.electrumx.stop()
        self.bitcoind.stop()

    def get_reduced_network_definition(self, network_definition):
        """
        Checks, if network definition was correct and extracts relevant
        information.

        The network size is reduced depending on self.node_limit.
        :param network_definition: dict
        :return: dict
        """
        network = {}
        for node_name, node_instance in network_definition.nodes.items():
            node = {}
            if node_name <= self.node_limit:
                # figure out which lightning daemon should be used
                daemon = node_instance.get('daemon')
                if daemon in AVAILABLE_DAEMONS.keys():
                    pass
                else:
                    daemon = 'lnd'

                node['daemon'] = daemon
                node['grpc_port'] = node_instance['grpc_port']
                node['rest_port'] = node_instance['rest_port']
                node['port'] = node_instance['port']
                node['base_fee_msat'] = node_instance['base_fee_msat']
                node['fee_rate'] = node_instance['fee_rate']
                channels = {}
                for channel, channel_properties in \
                        node_instance['channels'].items():
                    if channel_properties['to'] <= self.node_limit:
                        channels[channel] = channel_properties
                node['channels'] = channels
                network[node_name] = node
        return network

    def print_cli_commands(self):
        """
        Prints commands for controlling the lightning nodes directly from
        the shell.
        """
        logger.info('lncli commands:')
        for node_name, node_instance in self.ln_nodes.items():
            node_instance.print_rpc_command()

    def nodes_start(self):
        """
        Starts all LN nodes.
        """
        for node_name, node_instance in self.ln_nodes.items():
            node_instance.start(from_scratch=self.from_scratch)

    def nodes_stop(self):
        """
        Stops all LN nodes.
        """
        for node_name, node_instance in self.ln_nodes.items():
            try:
                node_instance.stop()
            except Exception as e:
                print(e)

    def nodes_set_pubkeys(self):
        """
        Tells the LN nodes to set their node pub keys.

        This can only be done after the RPC is up.
        """
        for node_name, node_instance in self.ln_nodes.items():
            node_instance.set_node_pubkey()

    def nodes_get_addresses(self):
        """
        Generates addresses in LN nodes' wallets.

        :return: list of str
            List of addresses.
        """
        addresses = []
        for node_name, node_instance in self.ln_nodes.items():
            address = node_instance.getaddress()
            logger.info("%s: %s", node_name, address)
            addresses.append(address)
        return addresses

    def nodes_connect_open_channels(self):
        """
        Connects LN nodes and opens channels between them.
        """
        for node_name, node_instance in self.ln_nodes.items():
            for channel, channel_data in \
                    self.network_definition[node_name]['channels'].items():

                # connect nodes
                node_to_connect = channel_data['to']
                if node_to_connect > self.node_limit:
                    continue

                # prepare connection info
                node_pubkey = self.ln_nodes[node_to_connect].pubkey
                node_port = self.ln_nodes[node_to_connect].lnport
                node_host = 'localhost:{}'.format(node_port)

                # prepare channel info
                capacity = int(channel_data['capacity'])
                total_relative = (channel_data['ratio_local'] +
                                  channel_data['ratio_remote'])
                local_relative = \
                    float(channel_data['ratio_local']) / total_relative
                remote_relative = \
                    float(channel_data['ratio_remote']) / \
                    total_relative
                remote_sat = int(capacity * remote_relative)
                local_sat = int(capacity * local_relative)

                # connect and open channel
                funding_txid = node_instance.connect_and_openchannel(
                    node_pubkey,
                    node_host,
                    capacity,
                    remote_sat
                )
                self.bitcoind.mine_blocks(3)
                # TODO: wait for lnd wallet to be synced
                time.sleep(1)

                # save funding txid, to later on get a channel mapping
                self.channel_mapping[channel]['funding_txid'] = funding_txid

            # finalize channel creation
            self.bitcoind.mine_blocks(3)
            time.sleep(WAIT_AFTER_MINING_THREE)

    def nodes_print_info(self):
        """
        Prints out essential information about the state of a node.
        """
        for node_name, node_instance in self.ln_nodes.items():
            info = node_instance.getinfo()
            logger.debug(
                f"{node_name}:\n{pformat(info, indent=4)}"
            )

    def nodes_fill_wallets(self):
        """
        Funds LN nodes' wallets.
        """
        addresses = self.nodes_get_addresses()
        self.bitcoind.sendtoaddresses(addresses, amount=1)
        self.bitcoind.mine_blocks(6)

    def nodes_set_fees(self):
        """
        Updates fees for each node.
        """
        for node_name, node_instance in self.ln_nodes.items():
            node_definition = self.network_definition[node_name]
            logger.info("%s: Update node policy.", node_name)
            node_instance.updatechanpolicy(
                node_definition['base_fee_msat'],
                node_definition['fee_rate'],
            )
        # TODO: set fees on a per channel basis

    def determine_node_mapping(self):
        """
        Updates the node mappings.

        Sets the mappings (dicts)
            self.node_mapping: node name (e.g. 'A') -> node pub key
            self.node_mapping_inverse: node pub key -> node name (e.g. 'A')
        """
        self.nodes_set_pubkeys()
        for node_name, node_instance in self.ln_nodes.items():
            self.node_mapping[node_name] = node_instance.pubkey
        self.node_mapping_inverse = {
            p: n for n, p in self.node_mapping.items()}

    def determine_channel_mapping(self):
        """
        Updates the channel mappings.

        Sets the mappings (dicts)
            self.channel_mapping: channel number (e.g. 3) ->
                channel_id, funding_txid, channel_point
            self.channel_mapping_inverse: channel_id ->
                channel number (e.g. 3)
        """
        logger.debug("NET: determine channel mapping")

        # create a temporary mapping from funding tx to the channel number
        map_ftx_to_cn = {
            f['funding_txid']: i for i, f in self.channel_mapping.items()}

        for node_name, node_instance in self.ln_nodes.items():
            # electrum takes a bit longer to reflect the channel state
            if isinstance(node_instance, Electrum):
                node_instance.wait_all_channels_open()

            channel_states = node_instance.listchannels()

            for channel in channel_states:
                # map channel id to funding transaction
                channel_number = map_ftx_to_cn[channel.funding_txid]
                self.channel_mapping[channel_number]['channel_id'] = \
                    int(channel.channel_id)
                self.channel_mapping[channel_number]['channel_point'] = \
                    int(channel.outpoint)

            logger.info(f"{node_name}: channel states {channel_states}")

        # also set the inverse mapping
        self.channel_mapping_inverse = {
            p['channel_id']: n for n, p in self.channel_mapping.items()}

        logger.debug(
            f"NET: channel mapping\n{pformat(self.channel_mapping, indent=4)}")

    def assemble_graph(self):
        """
        Gives a representation of the state of the LN network.

        The graph cannot be fetched from describegraph, as each node only has
        a local view of the network, therefore we need to ask each node
        about its open channels and gather the information.

        :return: dict
        """
        graph = {}

        for node_name, node_instance in self.ln_nodes.items():
            edges = {}
            channel_info = node_instance.listchannels()
            logger.debug(f"NET: channel info{channel_info}")
            for channel in channel_info:
                edge = {
                    'remote_name':
                        self.node_mapping_inverse[channel.remote_pubkey],
                    'capacity': int(channel.capacity),
                    'local_balance': int(channel.local_balance),
                    'remote_balance': int(channel.remote_balance),
                    'commit_fee': int(channel.commit_fee),
                    'initiator': channel.initiator,
                }
                # TODO: extend with more properties
                edges[self.channel_mapping_inverse[
                    int(channel.channel_id)]] = edge
            graph[node_name] = edges

        # TODO: extend with feereport properties
        return graph

    def save_mappings(self):
        """
        Pickles the node and channel mappings.
        """
        with open(self.node_mapping_path, 'wb') as f:
            pickle.dump(self.node_mapping, f, protocol=pickle.HIGHEST_PROTOCOL)
        with open(self.channel_mapping_path, 'wb') as f:
            pickle.dump(
                self.channel_mapping, f, protocol=pickle.HIGHEST_PROTOCOL)

    def read_mappings(self):
        """
        Loads the node and channel mappings.
        """
        with open(self.node_mapping_path, 'rb') as f:
            self.node_mapping = pickle.load(f)
            self.node_mapping_inverse = {
                p: n for n, p in self.node_mapping.items()}
        with open(self.channel_mapping_path, 'rb') as f:
            self.channel_mapping = pickle.load(f)
            self.channel_mapping_inverse = {
                p['channel_id']: n for n, p in self.channel_mapping.items()}

    def master_node_graph_view(self):
        """
        Prints the graph view of the master node.
        """
        chan_infos = self.master_node.describegraph()
        logger.debug(f'NET: {pformat(chan_infos)}')
        # extract relevant information from graph and map pub keys and
        # channel ids to human readable identifiers
        channel_list = []
        for c in chan_infos:
            # handle different master node instances differently due to
            # different outputs of their graph methods
            node_name_1 = self.node_mapping_inverse[c.node1_key]
            node_name_2 = self.node_mapping_inverse[c.node2_key]
            channel_number = self.channel_mapping_inverse[c.channel_id]
            if node_name_1 < node_name_2:
                channel_list.append((node_name_1, node_name_2, channel_number))
            else:
                channel_list.append((node_name_2, node_name_1, channel_number))
        channel_list.sort(key=lambda x: (x[0], x[2]))

        logger.info("Graph view of master node:")
        for channel in channel_list:
            logger.info("{} -> {} (channel #{})".format(
                channel[0], channel[1], channel[2]))

    def master_node_stop_and_start(self):
        self.master_node.stop()
        self.master_node.start(from_scratch=False)

    def master_node_print_networkinfo(self):
        logger.info("Master node info:")
        logger.info(format_dict(self.master_node.getnetworkinfo()))


if __name__ == '__main__':
    import logging.config

    logging.config.dictConfig(logger_config)
    logger.level = logging.INFO

    testnet = Network(
        network_definition_location='star_ring', node_limit='I',
        from_scratch=True)

    try:
        testnet.run_nocleanup()
        logger.info(format_dict(testnet.channel_mapping))
        logger.info(format_dict(testnet.node_mapping))
        logger.info(format_dict(testnet.assemble_graph()))
    finally:
        testnet.stop_components()



"""
Module can be used to check if graphs are defined in the correct convention.
"""

from typing import Dict


def graph_test(nodes):
    """
    Tests if a graph was defined in a certain convention.

    :param nodes: nodes definition
    :type nodes: dict
    """
    channel_numbers = sorted(get_channel_numbers(nodes))

    test_node_names_alphabetical(nodes)
    test_channel_numbers_unique(channel_numbers)
    test_ports(nodes)
    test_allowed_node_implementations(nodes)


def test_allowed_node_implementations(nodes: Dict[str, Dict]):
    """
    Tests if the daemon fields in the node definitions is from the
    supported set of daemons.
    """
    allowed_nodes = {None, 'electrum', 'lnd'}  # None is lnd default
    present_nodes = set()
    for n in nodes.values():
        present_nodes.add(n.get('daemon'))
    for pn in present_nodes:
        if pn not in allowed_nodes:
            raise ValueError(
                f"Error in daemon field of graph definition, "
                f"should be one of {allowed_nodes}, is {pn}.")



def get_channel_numbers(nodes):
    """
    Extracts channel numbers from graph definition.

    :param nodes: nodes definintion
    :type nodes: dict
    :return: channel numbers
    :rtype: list(int)
    """
    channels = []
    for node_name, node_data in nodes.items():
        channels.extend(node_data['channels'].keys())
    return channels


def test_node_names_alphabetical(nodes):
    """
    Tests if the names of the nodes were defined in an alpabetical increasing
    order.

    :param nodes: nodes definition
    :type nodes: dict

    :return: True if test matches expectation
    :rtype: bool
    """
    node_names = nodes.keys()
    number_nodes = len(node_names)
    alphabet = [str(chr(i)) for i in range(65, 91)]
    node_names_should = alphabet[:number_nodes]
    assert node_names_should == list(node_names),\
        "Node names do not follow convention A, B, C, ..."


def test_channel_numbers_unique(channel_numbers):
    assert len(set(channel_numbers)) == len(channel_numbers), \
        'Channel numbers are not unique.'


def get_ports(nodes):
    """
    Extracts ports from nodes definition.

    :param nodes: nodes definition
    :type nodes: dict
    :return: lnd ports, grpc ports, rest ports
    :rtype: list, list, list
    """
    ports = []
    grpc_ports = []
    rest_ports = []
    for node_name, node_data in nodes.items():
        ports.append(node_data['port'])
        grpc_ports.append(node_data['grpc_port'])
        rest_ports.append(node_data['port'])
    return ports, grpc_ports, rest_ports


def test_ports(nodes):
    ports, grpc_ports, rest_ports = get_ports(nodes)

    assert len(nodes) == len(ports)
    assert len(nodes) == len(grpc_ports)
    assert len(nodes) == len(rest_ports)

    assert len(ports) == len(set(ports))
    assert len(grpc_ports) == len(set(grpc_ports))
    assert len(rest_ports) == len(set(rest_ports))


if __name__ == '__main__':
    from lnregtest.network_definitions.default import nodes as star_ring_nodes
    graph_test(star_ring_nodes)



"""
Integration tests for rebalancing of channels.
"""
import time

from lndmanage import settings
from lndmanage.lib.listchannels import ListChannels
from lndmanage.lib.rebalance import Rebalancer
from lndmanage.lib.ln_utilities import channel_unbalancedness_and_commit_fee
from lndmanage.lib.exceptions import RebalanceCandidatesExhausted
from test.testnetwork import TestNetwork

from test.testing_common import (
    lndmanage_home,
    test_graphs_paths,
    SLEEP_SEC_AFTER_REBALANCING)

import logging.config
settings.set_lndmanage_home_dir(lndmanage_home)
logging.config.dictConfig(settings.logger_config)
logger = logging.getLogger()
logger.setLevel(logging.DEBUG)
logger.handlers[0].setLevel(logging.DEBUG)


class RebalanceTest(TestNetwork):
    """
    Implements an abstract testing class for channel rebalancing.
    """
    def rebalance_and_check(self, test_channel_number, target,
                            allow_unbalancing, places=5):
        """
        Test function for rebalancing to a specific target unbalancedness and
        asserts afterwards that the target was reached.

        :param test_channel_number: channel id
        :type test_channel_number: int
        :param target: unbalancedness target
        :type target: float
        :param allow_unbalancing: if unbalancing should be allowed
        :type allow_unbalancing: bool
        :param places: accuracy of the comparison between expected and tested
            values
        :type places: int
        """
        rebalancer = Rebalancer(
            self.lndnode,
            max_effective_fee_rate=50,
            budget_sat=20
        )

        channel_id = self.testnet.channel_mapping[
            test_channel_number]['channel_id']

        try:
            fees_msat = rebalancer.rebalance(
                channel_id,
                dry=False,
                chunksize=1.0,
                target=target,
                allow_unbalancing=allow_unbalancing
            )
        except Exception as e:
            raise e

        # sleep a bit to let LNDs update their balances
        time.sleep(SLEEP_SEC_AFTER_REBALANCING)

        # check if graph has the desired channel balances
        graph = self.testnet.assemble_graph()
        channel_data = graph['A'][test_channel_number]
        listchannels = ListChannels(self.lndnode)
        listchannels.print_all_channels('rev_alias')

        channel_unbalancedness, _ = channel_unbalancedness_and_commit_fee(
            channel_data['local_balance'],
            channel_data['capacity'],
            channel_data['commit_fee'],
            channel_data['initiator']
        )

        self.assertAlmostEqual(
            target, channel_unbalancedness, places=places)

        return fees_msat

    def graph_test(self):
        """
        graph_test should be implemented by each subclass test and check,
        whether the test graph has the correct shape.
        """
        raise NotImplementedError


class TestLiquidRebalance(RebalanceTest):
    network_definition = test_graphs_paths['star_ring_3_liquid']

    # channels:
    # A -> B (channel #1)
    # A -> C (channel #2)
    # A -> D (channel #6)
    # B -> C (channel #3)
    # B -> D (channel #4)
    # C -> D (channel #5)

    def graph_test(self):
        self.assertEqual(4, self.master_node_networkinfo['num_nodes'])
        self.assertEqual(6, self.master_node_networkinfo['num_channels'])

    def test_rebalance_channel_6(self):
        test_channel_number = 6
        self.rebalance_and_check(test_channel_number, 0.0, False)

    def test_small_positive_target_channel_6(self):
        test_channel_number = 6
        self.rebalance_and_check(test_channel_number, 0.2, False)

    def test_large_positive_channel_6(self):
        test_channel_number = 6
        self.rebalance_and_check(test_channel_number, 0.8, False)

    def test_small_negative_target_channel_6_fail(self):
        # this test should fail when unbalancing is not allowed, as it would
        # unbalance another channel if the full target would be accounted for
        test_channel_number = 6
        self.assertRaises(
            RebalanceCandidatesExhausted,
            self.rebalance_and_check, test_channel_number, -0.2, False)

    def test_small_negative_target_channel_6_succeed(self):
        # this test should fail when unbalancing is not allowed, as it would
        # unbalance another channel if the full target would be accounted for
        test_channel_number = 6
        self.rebalance_and_check(test_channel_number, -0.2, True)

    def test_rebalance_channel_1(self):
        test_channel_number = 1
        self.rebalance_and_check(test_channel_number, 0.0, False)

    def test_rebalance_channel_2(self):
        test_channel_number = 2
        self.rebalance_and_check(test_channel_number, 0.0, False, places=1)

    def test_shuffle_arround(self):
        """
        Shuffles sat around in channel 6.
        """
        first_target_amount = -0.1
        second_target_amount = 0.1
        test_channel_number = 6

        self.rebalance_and_check(
            test_channel_number, first_target_amount, True)
        self.rebalance_and_check(
            test_channel_number, second_target_amount, True)


class TestUnbalancedRebalance(RebalanceTest):
    network_definition = test_graphs_paths['star_ring_4_unbalanced']

    # channels:
    # A -> B (channel #1)
    # A -> C (channel #2)
    # A -> D (channel #3)
    # A -> E (channel #4)

    def graph_test(self):
        self.assertEqual(5, self.master_node_networkinfo['num_nodes'])
        self.assertEqual(10, self.master_node_networkinfo['num_channels'])

    def test_rebalance_channel_1(self):
        """tests multiple rebalance of one channel"""
        test_channel_number = 1
        # TODO: find out why not exact rebalancing target is reached
        print(self.rebalance_and_check(
            test_channel_number, -0.05, False, places=1))


class TestIlliquidRebalance(RebalanceTest):
    network_definition = test_graphs_paths['star_ring_4_illiquid']

    # channels:
    # A -> B (channel #1)
    # A -> C (channel #2)
    # A -> D (channel #3)
    # A -> E (channel #4)

    def graph_test(self):
        self.assertEqual(5, self.master_node_networkinfo['num_nodes'])
        self.assertEqual(10, self.master_node_networkinfo['num_channels'])

    def test_rebalance_channel_1(self):
        """
        Tests multiple payment attempt rebalancing.
        """
        test_channel_number = 1
        # TODO: find out why not exact rebalancing target is reached
        fees_msat = self.rebalance_and_check(
            test_channel_number, -0.05, False, places=1)
        self.assertEqual(2575, fees_msat)

    def test_rebalance_channel_1_fail(self):
        """
        Tests if there are no rebalance candidates, because the target
        requested doesn't match with the other channels.
        """
        test_channel_number = 1
        self.assertRaises(
            RebalanceCandidatesExhausted, self.rebalance_and_check,
            test_channel_number, 0.3, False, places=1
        )



from lndmanage.lib.rating import ChannelRater
from lndmanage.lib.exceptions import RouteWithTooSmallCapacity, NoRoute
from lndmanage.lib.pathfinding import ksp_discard_high_cost_paths
from lndmanage import settings

import logging
logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())


def calculate_fees_on_policy(amt_msat, policy):
    return policy['fee_base_msat'] + amt_msat * policy['fee_rate_milli_msat'] // 1000000


class Route(object):
    """
    Deals with the onion route construction from list of channels. Calculates fees and cltvs.

    :param node: :class:`lib.node.Node` instance
    :param channel_hops: list of chan_ids along which the route shall be constructed
    :param node_dest: pub_key of destination node
    :param amt_msat: amount to send in msat
    """

    def __init__(self, node, channel_hops, node_dest, amt_msat):
        self.node = node
        self.blockheight = node.blockheight
        logger.debug(f"Blockheight: {self.blockheight}")
        self.channel_hops = channel_hops
        self._hops = []
        self._node_hops = [node_dest]
        forward_msat = amt_msat
        final_cltv = 144
        fees_msat_container = [0]
        cltv_delta = [0]
        node_right = node_dest
        node_left = None
        policy = None

        logger.debug("Route construction starting.")
        # hops are traversed in backwards direction to accumulate fees and cltvs
        for ichannel, channel_id in enumerate(reversed(channel_hops)):
            channel_data = self.node.network.edges[channel_id]

            if amt_msat // 1000 > channel_data['capacity']:
                logger.debug(f"Discovered a channel {channel_id} with too small capacity.")
                raise RouteWithTooSmallCapacity(f"Amount too large for channel.")

            if node_right == channel_data['node2_pub']:
                try:
                    policy = channel_data['node1_policy']
                    node_left = channel_data['node1_pub']
                except KeyError:
                    logger.exception(f"No channel {channel_data}")
            else:
                policy = channel_data['node2_policy']
                node_left = channel_data['node2_pub']

            self._node_hops.append(node_left)

            logger.debug(f"Route from {node_left}")
            logger.debug(f"        to {node_right}")
            logger.debug(f"Policy of forwarding node: {policy}")

            fees_msat = policy['fee_base_msat'] + policy['fee_rate_milli_msat'] * forward_msat // 1000000
            forward_msat = amt_msat + sum(fees_msat_container[:ichannel])
            fees_msat_container.append(fees_msat)

            logger.debug(f"Hop: {len(channel_hops) - ichannel}")
            logger.debug(f"     Fees: {fees_msat}")
            logger.debug(f"     Fees container {fees_msat_container}")
            logger.debug(f"     Forward: {forward_msat}")

            self._hops.append({
                'chan_id': channel_data['channel_id'],
                'chan_capacity': channel_data['capacity'],
                'amt_to_forward': forward_msat // 1000,
                'fee': fees_msat_container[-2] // 1000,
                'expiry': self.blockheight + final_cltv + sum(cltv_delta[:ichannel]),
                'amt_to_forward_msat': forward_msat,
                'fee_msat': fees_msat_container[-2],
            })

            cltv_delta.append(policy['time_lock_delta'])
            node_right = node_left

        self.hops = list(reversed(self._hops))
        self.node_hops = list(reversed(self._node_hops))
        self.total_amt_msat = amt_msat + sum(fees_msat_container[:-1])
        self.total_fee_msat = sum(fees_msat_container[:-1])
        self.total_time_lock = sum(cltv_delta[:-1]) + self.blockheight + final_cltv

    def _debug_route(self):
        """
        Prints detailed information of the route.
        """
        logger.debug("Debug route:")
        for h in self.hops:
            logger.debug(f"c:{h['chan_id']} a:{h['amt_to_forward']} f:{h['fee_msat']}"
                         f" c:{h['expiry'] - self.node.blockheight}")
            channel_info = self.node.network.edges[h['chan_id']]
            node1 = channel_info['node1_pub']
            node2 = channel_info['node2_pub']
            logger.debug(f"{node1[:5]}, {channel_info['node1_policy']}")
            logger.debug(f"{node2[:5]}, {channel_info['node2_policy']}")
        logger.debug(f"tl:{self.total_time_lock} ta:{self.total_amt_msat} tf:{self.total_fee_msat}")


class Router(object):
    """
    Contains utilities for constructing routes.

    :param node: :class:`lib.node.Node` instance
    """

    def __init__(self, node):
        self.node = node
        self.channel_rater = ChannelRater()

    def _node_route_to_channel_route(self, node_route, amt_msat):
        """
        Takes a route in terms of a list of nodes and translates it into a list of channels.

        :param node_route: list of pubkeys
        :param amt_msat: amount to send in sat
        :return: list of channel_ids
        """
        channels = []
        for p in range(len(node_route) - 1):
            channels.append(
                self._determine_cheapest_channel_between_two_nodes(
                    node_route[p], node_route[p + 1], amt_msat)[1])
        return channels

    def get_routes_from_to_nodes(self, node_from, node_to, amt_msat, number_of_routes=10):
        """
        Determines number_of_routes shortest paths between node_from and node_to for an amount of amt_msat.

        :param node_from: pubkey
        :param node_to: pubkey
        :param amt_msat: amount to send in msat
        :param number_of_routes: int
        :return: number_of_routes lists of node pubkeys
        """
        self.channel_rater.bad_nodes.append(self.node.pub_key)  # excludes self-loops

        weight_function = lambda v, u, e: self.channel_rater.node_to_node_weight(v, u, e, amt_msat)
        routes = ksp_discard_high_cost_paths(
            self.node.network.graph, node_from, node_to,
            num_k=number_of_routes, weight=weight_function)

        if not routes:
            raise NoRoute

        return routes

    def _determine_cheapest_channel_between_two_nodes(self, node_from, node_to, amt_msat):
        """
        Determines the cheapest channel between nodes node_from and node_to for an amount of amt_msat.

        :param node_from: pubkey
        :param node_to: pubkey
        :param amt_msat: amount to send in msat
        :return: channel_id
        """
        number_edges = self.node.network.graph.number_of_edges(node_from, node_to)
        channels_with_calculated_fees = []
        for n in range(number_edges):
            edge = self.node.network.graph.get_edge_data(node_from, node_to, n)
            fees = self.channel_rater.channel_weight(edge, amt_msat)
            channels_with_calculated_fees.append([fees, edge['channel_id']])
        sorted_channels = sorted(channels_with_calculated_fees, key=lambda k: k[0])
        return sorted_channels[0]

    def _determine_cheapest_fees_between_two_nodes(self, node_from, node_to, amt):
        return self._determine_cheapest_channel_between_two_nodes(node_from, node_to, amt)[0]

    def get_route_channel_hops_from_to_node_internal(self, source_pubkey, target_pubkey, amt_msat):
        """
        Find routes internally, using networkx to construct a route from a source node to a target node.

        :param source_pubkey: str
        :param target_pubkey: str
        :param amt_msat: int
        :return:
        """
        logger.debug(f"Internal route finding:")
        logger.debug(f"from {source_pubkey}")
        logger.debug(f"  to {target_pubkey}")

        node_routes = self.get_routes_from_to_nodes(
            source_pubkey, target_pubkey, amt_msat, number_of_routes=1)

        hops = self._node_route_to_channel_route(node_routes[0], amt_msat)

        # logger.debug(f"(Intermediate) route as channel hops: {hops}")
        return [hops]

    def get_route_channel_hops_from_to_node_external(
            self, source_pubkey, target_pubkey, amt_msat, use_mc=False):
        """
        Find routes externally (relying on the node api) to construct a route
        from a source node to a target node.

        :param source_pubkey: source public key
        :type source_pubkey: str
        :param target_pubkey: target public key
        :type target_pubkey: str
        :param amt_msat: amount to send in msat
        :type amt_msat: int
        :param use_mc: true if mission control based pathfinding is used
        :type use_mc: bool
        :return: list of hops
        :rtype: list[list[int]]
        """
        logger.debug(f"External pathfinding, using mission control: {use_mc}.")
        logger.debug(f"from {source_pubkey}")
        logger.debug(f"  to {target_pubkey}")
        ignored_nodes = self.channel_rater.bad_nodes

        # we don't need to give blacklisted channels to the queryroute command
        # as all of this is done by mission control
        if use_mc:
            ignored_channels = {}
        else:
            ignored_channels = self.channel_rater.bad_channels

        hops = self.node.queryroute_external(
            source_pubkey, target_pubkey, amt_msat,
            ignored_channels=ignored_channels,
            ignored_nodes=ignored_nodes,
            use_mc=use_mc,
        )

        return [hops]

    def get_routes_for_rebalancing(
            self, chan_id_from, chan_id_to, amt_msat, method='external'):
        """
        Calculates several routes for channel_id_from to chan_id_to
        and optimizes for fees for an amount amt.

        :param chan_id_from: short channel id of the from node
        :type chan_id_from: int
        :param chan_id_to: short channel id of the to node
        :type chan_id_to: int
        :param amt_msat: payment amount in msat
        :type amt_msat: int
        :param method: specifies if 'internal', or 'external'
               method of route computation should be used
        :type method: string
        :return: list of :class:`lib.routing.Route` instances
        :rtype: list[lndmanage.lib.routing.Route]
        """

        try:
            channel_from = self.node.network.edges[chan_id_from]
            channel_to = self.node.network.edges[chan_id_to]
        except KeyError:
            logger.exception(
                "Channel was not found in network graph, but is present in "
                "listchannels. Channel needs 6 confirmations to be usable.")
            raise NoRoute

        this_node = self.node.pub_key

        # find the correct node_pubkeys between which we want to route
        # fist hop:start-end ----- last hop: start-end
        if channel_from['node1_pub'] == this_node:
            first_hop_end = channel_from['node2_pub']
        else:
            first_hop_end = channel_from['node1_pub']

        if channel_to['node1_pub'] == this_node:
            last_hop_start = channel_to['node2_pub']
        else:
            last_hop_start = channel_to['node1_pub']

        # determine inner channel hops
        # internal method uses networkx dijkstra,
        # this is more independent, but slower
        if method == 'internal':
            routes_channel_hops = \
                self.get_route_channel_hops_from_to_node_internal(
                first_hop_end, last_hop_start, amt_msat)
        # rely on external pathfinding with internal blacklisting
        elif method == 'external':
            routes_channel_hops = \
                self.get_route_channel_hops_from_to_node_external(
                first_hop_end, last_hop_start, amt_msat, use_mc=False)
        # rely on external pathfinding using mission control
        elif method == 'external-mc':
            routes_channel_hops = \
                self.get_route_channel_hops_from_to_node_external(
                first_hop_end, last_hop_start, amt_msat, use_mc=True)
        else:
            raise ValueError(
                f"Method must be either internal, external or external-mc, "
                f"is {method}.")

        # pre- and append the outgoing and incoming channels to the route
        routes_channel_hops_final = []
        for r in routes_channel_hops:
            r.insert(0, chan_id_from)
            r.append(chan_id_to)
            logger.debug("Channel hops:")
            logger.debug(r)
            routes_channel_hops_final.append(r)

        # initialize Route objects with appropriate fees and expiries
        routes = []
        for h in routes_channel_hops:
            try:
                route = Route(self.node, h, this_node, amt_msat)
                routes.append(route)
            except RouteWithTooSmallCapacity:
                continue
        return routes


if __name__ == '__main__':
    import logging.config
    logging.config.dictConfig(settings.logger_config)
    from lndmanage.lib.node import LndNode
    nd = LndNode()



"""
Network components for the bitcoin/lightning regtest network.
"""
import asyncio
from abc import ABC, abstractmethod
from functools import wraps
import subprocess
import shutil
import time
import os
import os.path
import logging
import pathlib
import threading
from typing import Optional, List, NamedTuple
import re

from .node_config_templates import (
    lnd_config_template,
    bitcoind_config_template
)
from lnregtest.lib.utils import decode_byte_string_to_dict_or_str, convert_short_channel_id_to_channel_id, bfh

from lnregtest.lib.common import WAIT_SYNC_BITCOIND, WAIT_SYNC_ELECTRUMX

logger = logging.getLogger(__name__)
logger.addHandler(logging.NullHandler())


class ChannelState(NamedTuple):
    capacity: int
    channel_id: int
    commit_fee: int
    funding_txid: str
    initiator: bool
    local_balance: int
    outpoint: int
    remote_balance: int
    remote_pubkey: str
    state: str


class Graph(NamedTuple):
    funding_txid: str
    outpoint: int
    state: str


class ChannelInfo(NamedTuple):
    channel_id: int
    node1_key: str
    node2_key: str


def log_subprocess_output(pipe, prefix=''):
    for line in iter(pipe.readline, b''):  # b'\n'-separated lines
        if prefix:
            logger.debug(f'{prefix}: {str(line)}')
        else:
            logger.debug(f'{str(line)}')


class Bitcoind(object):
    """
    Bitcoind node abstraction.
    """

    def __init__(self, nodedata_folder: str,
                 binary_folder: Optional[str] = None):
        """
        :param nodedata_folder: str:
            absolute path to folder, where node data is stored
        :param binary_folder: str:
            absolute path where bitcoind/bitcoin-cli can be found, else
            binaries are taken from $PATH
        """
        self.nodedata_folder = nodedata_folder
        self.bitcoind_config_file = \
            os.path.join(self.nodedata_folder, 'bitcoin/bitcoin.conf')
        self.bitcoind_data_dir = os.path.join(self.nodedata_folder, 'bitcoin')

        # take binaries from path, if no binary folder is given
        if binary_folder is None:
            binary_folder = ''
        self.bitcoind_binary = os.path.join(binary_folder, 'bitcoind')
        self.bitcoincli_binary = os.path.join(binary_folder, 'bitcoin-cli')

        if shutil.which(self.bitcoind_binary) is None:
            raise FileNotFoundError(
                f"bitcoind executable not found: {self.bitcoind_binary}")
        if shutil.which(self.bitcoincli_binary) is None:
            raise FileNotFoundError(
                f"bitcoin-cli executable not found: {self.bitcoincli_binary}")

        self.bitcoind_process = None

    def start(self, from_scratch=True):
        """
        Gets bitcoind going.

        `from_scratch` determines, if the network should be built from a clean
        state.

        :param from_scratch: bool
        """

        # prepare data directories
        if from_scratch:
            self.clear_directory()
            self.setup_bitcoinddir()
        else:
            if not os.path.isdir(self.bitcoind_data_dir):
                raise FileNotFoundError(
                    'Bitcoind data directory not found '
                    '(from_scratch = False).')

        # start and wait
        self.start_bitcoind_process()
        self.block_until_started()
        if from_scratch:
            self.create_wallet()
        logger.info("BTC: Bitcoind started.")

    def setup_bitcoinddir(self):
        """
        Sets up the bitcoind data folder.
        """
        pathlib.Path(self.bitcoind_data_dir).mkdir(parents=True, exist_ok=True)
        config = bitcoind_config_template

        with open(self.bitcoind_config_file, 'w') as f:
            f.write(config)

    def create_wallet(self):
        """
        Creates a new wallet for bitcoind.
        """
        self.bitcoincli(['createwallet', 'default'])

    def start_bitcoind_process(self):
        """
        Starts a bitcoind subprocess.
        """
        command = [self.bitcoind_binary,
                   '-datadir=' + self.bitcoind_data_dir]

        logger.info("BTC: Starting bitcoind.")
        logger.info(' '.join(command))
        self.bitcoind_process = subprocess.Popen(
            command,
            stdout=subprocess.DEVNULL,
            stderr=subprocess.DEVNULL,
            start_new_session=True,
        )

    def stop(self):
        self.bitcoincli(['stop'])
        logger.info("BTC: stopped bitcoind")
        # should wait for the process to really stop
        self.bitcoind_process.communicate()

    def block_until_started(self):
        """
        Checks, if bitcoind has started by looking at the cli response and
        blocks if it doesn't get proper response.
        """
        while True:
            result = self.bitcoincli(['getblockchaininfo'])
            started = True if result.returncode == 0 else False
            if started:
                break
            time.sleep(WAIT_SYNC_BITCOIND)

    def clear_directory(self):
        """
        Deletes the bitcoind regtest folder.
        """
        logger.debug("BTC: Cleaning up bitcoin data directory.")
        try:
            shutil.rmtree(self.bitcoind_data_dir)
        except FileNotFoundError as e:
            logger.debug("BTC: Directory already clean. %s", e)

    def bitcoincli(self, command):
        """
        Invokes the bitcoin-cli command line interface.

        :param command: list, contains CLI parameters
        :return: subprocess
        """
        cmd = [self.bitcoincli_binary, '-datadir=' +
               self.bitcoind_data_dir] + command
        logger.debug(' '.join(cmd))
        proc = subprocess.run(
            cmd,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )

        # Handle errors of the bitcoin-cli call. Errors are expected
        # to occur for getblockchaininfo when the daemon is not ready to
        # take calls, therefore we exclude this case here.
        if command[0] != 'getblockchaininfo' and proc.returncode != 0:
            raise ChildProcessError(proc.stderr)

        return proc

    def getblockchaininfo(self):
        proc = self.bitcoincli(['getblockchaininfo'])
        return decode_byte_string_to_dict_or_str(proc.stdout)

    def get_blockheight(self):
        blockchaininfo = self.getblockchaininfo()
        return blockchaininfo['blocks']

    def mine_blocks(self, number_of_blocks):
        """
        Mines some blocks.

        :param number_of_blocks: int
        """
        logger.info("BTC: Mining %d blocks.", number_of_blocks)
        address = self.newaddress()
        command = list(['generatetoaddress', str(number_of_blocks), address])
        result = self.bitcoincli(command)
        logger.debug("BTC: Mined to address %s.", result.stdout)

    def fill_addresses(self, number_of_addresses):
        """
        Generates several addresses and attaches UTXOs to them.

        :param number_of_addresses: int
        """
        for _ in range(number_of_addresses):
            address = self.newaddress()
            command = list(['generatetoaddress', str(1), address])
            result = self.bitcoincli(command)
            logger.info("BTC: Mined to address %s.", result.stdout)

    def sendtoaddress(self, address, amount):
        """
        Sends funds to a given address.

        :param address: str
        :param amount: float: amount in BTC
        """
        proc = self.bitcoincli(['sendtoaddress', address, str(amount)])
        return decode_byte_string_to_dict_or_str(proc.stdout)

    def sendtoaddresses(self, addresses, amount):
        """
        Sends funds to given list of addresses.

        :param addresses: list of str
        :param amount: float: amount in BTC
        """
        logger.info("BTC: Sending funds to addresses.")
        for a in addresses:
            self.sendtoaddress(a, amount)

    def newaddress(self, address_type='bech32'):
        """
        Generates a new address.

        :param address_type: str: default is bech32
        :return: str: address
        """
        logger.debug("BTC: Getting new address.")
        command = ['getnewaddress', '', address_type]
        result = self.bitcoincli(command)
        address = result.stdout.strip()
        address = address.decode('utf-8')
        logger.debug("BTC: Generated new address %s.", address)
        return address

    def getbalances(self):
        """
        Gets confirmed and unconfirmed balances.

        :return: dict
        """
        logger.debug("BTC: Getting balances.")
        command = ['getbalances']
        result = self.bitcoincli(command)

        # convert into a dict
        balances = result.stdout.strip()
        balances = balances.decode('utf-8')
        balances = decode_byte_string_to_dict_or_str(balances)
        logger.debug("BTC: Generated new address %s.", balances)

        return balances


class ElectrumX(object):
    """
    Electrumx node abstraction.
    """

    def __init__(self, nodedata_folder: str,
                 binary_folder: Optional[str] = None):
        """
        :param nodedata_folder: str:
            absolute path to folder, where node data is stored
        :param binary_folder: str:
            absolute path where electrumx_server and electrumx_rpc can be found,
            else binaries are taken from $PATH
        """
        self.nodedata_folder = nodedata_folder
        self.db_directory = os.path.join(self.nodedata_folder, 'electrumx/db')
        self.services = "tcp://localhost:51001,rpc://localhost:8000"
        self.net = 'regtest'
        self.coin = 'BitcoinSegwit'
        self.daemon_url = "http://lnd:123456@localhost:18443"

        # export to environment variables
        os.environ['DB_DIRECTORY'] = self.db_directory
        os.environ['SERVICES'] = self.services
        os.environ['NET'] = self.net
        os.environ['COIN'] = self.coin
        os.environ['DAEMON_URL'] = self.daemon_url

        # take binaries from path, if no binary folder is given
        if binary_folder is None:
            binary_folder = ''
        self.server_binary = os.path.join(binary_folder, 'electrumx_server')
        self.rpc_binary = os.path.join(binary_folder, 'electrumx_rpc')

        if shutil.which(self.server_binary) is None:
            raise FileNotFoundError(
                f"electrumx_server executable not found: {self.server_binary}")
        if shutil.which(self.rpc_binary) is None:
            raise FileNotFoundError(
                f"electrumx_rpc executable not found: {self.rpc_binary}")

        self.server_process = None

    def start(self, from_scratch=True):
        """
        Gets electrumx going.

        `from_scratch` determines, if the network should be built from a clean
        state.

        :param from_scratch: bool
        """

        # prepare data directories
        if from_scratch:
            self.clear_directory()
            self.setup_db_dir()
        else:
            if not os.path.isdir(self.db_directory):
                raise FileNotFoundError(
                    'Electrumx database not found (from_scratch = False).')

        # start and wait
        self.start_process()
        self.block_until_started()
        logger.info("ETRX: Electrumx started.")

    def setup_db_dir(self):
        """
        Sets up the database folder.
        """
        pathlib.Path(self.db_directory).mkdir(parents=True, exist_ok=True)

    def start_process(self):
        """
        Starts an electrumx subprocess.
        """
        command = [self.server_binary]

        logger.info("ETRX: Starting electrumx.")
        self.server_process = subprocess.Popen(
            command,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            start_new_session=True,
        )
        # with self.server_process:
        #     log_subprocess_output(self.server_process.stdout)

    def stop(self):
        self.rpc(['stop'])
        logger.info("ETRX: stopping")
        # should wait for the process to really stop
        self.server_process.communicate()
        logger.info("ETRX: stopped")

    def block_until_started(self):
        """
        Checks, if bitcoind has started by looking at the cli response and
        blocks if it doesn't get proper response.
        """
        while True:
            result = self.rpc(['getinfo'])
            started = True if result.returncode == 0 else False
            if started:
                break
            time.sleep(WAIT_SYNC_BITCOIND)

    def clear_directory(self):
        """
        Deletes the database folder.
        """
        logger.debug("ETRX: Cleaning up the database directory.")
        try:
            shutil.rmtree(self.db_directory)
        except FileNotFoundError as e:
            logger.debug("ETRX: Directory already clean. %s", e)

    def rpc(self, command):
        """
        Invokes the electrumx rpc.

        :param command: list, contains CLI parameters
        :return: subprocess
        """
        cmd = [self.rpc_binary]
        cmd.extend(command)

        logger.debug(' '.join(cmd))
        proc = subprocess.run(
            cmd,
            stdout=subprocess.PIPE,
            stderr=subprocess.PIPE,
        )

        logger.debug(proc)

        return proc

    def getinfo(self):
        proc = self.rpc(['getinfo'])
        return decode_byte_string_to_dict_or_str(proc.stdout)


class LightningDaemon(ABC):
    # general attributes
    name: str
    lnport: int
    pubkey: Optional[str]
    node_properties: dict
    nodes_folder: str
    binary_folder: Optional[str]

    # lnd process
    server_process: Optional[subprocess.Popen]
    logs: List[str]
    logs_cond: threading.Condition
    running: bool
    thread: Optional[threading.Thread]
    server_binary: str
    rpc_command: List[str]

    # folders and binaries
    nodes_folder: str
    data_dir: str

    # fees
    base_fee_msat: int
    fee_rate: float

    def __init__(self, name, node_properties, nodes_folder,
                 binary_folder=None):
        """
        :param name: char: unique human readable identifier, e.g. A, B, ...
        :param node_properties: dict:
            as defined in network_definitions
        :param nodes_folder: str:
            absolute path to the node data folder
        :param binary_folder: str:
            absolute path to the binary folder, if not given, binaries are
            taken from $PATH
        """
        self.name = name
        self.lnport = node_properties['port']
        self.pubkey = None
        self.node_properties = node_properties

        # lnd process
        self.server_process = None
        self.logs = []
        self.logs_cond = threading.Condition(threading.RLock())
        self.running = False
        self.thread = None
        self.rpc_command = []

        # folders and binaries
        self.nodes_folder = nodes_folder
        self.data_dir = os.path.join(
            self.nodes_folder, 'lnnodes/' + self.name)

        # fees
        self.base_fee_msat = node_properties['base_fee_msat']
        self.fee_rate = node_properties['fee_rate']

    # following code was taken from github.com/cdecker/lightning-integration
    def tail(self):
        """Tail the stdout of the process and remember it.
        Stores the lines of output produced by the process in
        self.logs and signals that a new line was read so that it can
        be picked up by consumers.
        """
        try:
            for line in iter(self.server_process.stdout.readline, ''):
                if len(line) == 0:
                    break
                with self.logs_cond:
                    self.logs.append(str(line.rstrip()))
                    self.logs_cond.notifyAll()
        except ValueError:
            self.running = False

    def wait_for_log(self, regex, offset=1000, timeout=60):
        """Look for `regex` in the logs.
        We tail the stdout of the process and look for `regex`,
        starting from `offset` lines in the past. We fail if the
        timeout is exceeded or if the underlying process exits before
        the `regex` was found. The reason we start `offset` lines in
        the past is so that we can issue a command and not miss its
        effects.
        """
        logger.debug("%s: Waiting for '%s' in the logs", self.name, regex)
        ex = re.compile(regex)
        start_time = time.time()
        pos = max(len(self.logs) - offset, 0)
        initial_pos = len(self.logs)
        while True:
            if time.time() > start_time + timeout:
                print("Can't find {} in logs".format(regex))
                with self.logs_cond:
                    for i in range(initial_pos, len(self.logs)):
                        print("  " + self.logs[i])
                if self.is_in_log(regex):
                    print("(Was previously in logs!")
                raise TimeoutError(
                    'Unable to find "{}" in logs.'.format(regex))
            elif not self.running:
                print('Logs: {}'.format(self.logs))
                raise ValueError('Process died while waiting for logs')

            with self.logs_cond:
                if pos >= len(self.logs):
                    self.logs_cond.wait(1)
                    continue

                if ex.search(self.logs[pos]):
                    logging.debug("%s: Found '%s' in logs", self.name, regex)
                    return self.logs[pos]
                pos += 1

    def is_in_log(self, regex):
        """Look for `regex` in the logs."""
        ex = re.compile(regex)
        for log in self.logs:
            if ex.search(log):
                logging.debug("Found '%s' in logs", regex)
                return True

        logging.debug("Did not find '%s' in logs", regex)
        return False

    @abstractmethod
    def start(self) -> subprocess.Popen:
        pass

    @abstractmethod
    def stop(self):
        pass

    @abstractmethod
    def print_rpc_command(self):
        pass

    @abstractmethod
    def setup_datadir(self):
        pass

    @abstractmethod
    def clear_directory(self):
        pass

    @abstractmethod
    def listchannels(self) -> List[ChannelState]:
        pass

    def rpc(self, command: List[str]) -> (int, dict):
        """Invokes the rpc command line interface for the LN server.

        :param command: list of command line arguments
        :return:
            int: error code
            dict: generated from json response of cli
        """
        # make sure all arguments in list are str
        command = list(map(str, command))
        cmd = self.rpc_command + command
        logger.debug('%s: %s.', self.name, ' '.join(cmd))
        proc = subprocess.run(
            cmd,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
        )
        stdout_dict = decode_byte_string_to_dict_or_str(proc.stdout)
        logger.debug("%s: %s", self.name, stdout_dict)

        if proc.returncode:
            logger.error("%s: %s", self.name, proc.stderr)

        return proc.returncode, stdout_dict

    async def async_grpc(self, command: List[str]) -> (int, dict):
        """Async implementation of self.rpc."""

        command = list(map(str, command))
        cmd = ' '.join(self.rpc_command + command)
        logger.debug('%s: %s.', self.name, cmd)

        proc = await asyncio.create_subprocess_shell(
            cmd,
            stdout=asyncio.subprocess.PIPE,
            stderr=asyncio.subprocess.PIPE)

        stdout, stderr = await proc.communicate()

        stdout_dict = decode_byte_string_to_dict_or_str(stdout)
        logger.debug("%s: %s", self.name, stdout_dict)

        stderr_dict = decode_byte_string_to_dict_or_str(stderr)
        if proc.returncode:
            logger.error("%s: %s", self.name, stderr_dict)

        return proc.returncode, stdout_dict


class LND(LightningDaemon):
    def __init__(self, name, node_properties, nodes_folder,
                 binary_folder=None):
        super().__init__(name, node_properties, nodes_folder, binary_folder)

        # network definitions
        self._grpc_port = node_properties['grpc_port']
        self._rest_port = node_properties['rest_port']
        self._grpc_host = 'localhost:' + str(self._grpc_port)

        # take binaries from path, if no binary folder is given
        if binary_folder is None:
            binary_folder = ''
        self.server_binary = os.path.join(binary_folder, 'lnd')
        self.rpc_binary = os.path.join(binary_folder, 'lncli')

        # check if executables can be found
        if shutil.which(self.server_binary) is None:
            raise FileNotFoundError(
                f"lnd executable not found: {self.server_binary}")
        if shutil.which(self.rpc_binary) is None:
            raise FileNotFoundError(
                f"lncli executable not found: {self.rpc_binary}")

        # file paths
        self._cert_file = os.path.join(self.data_dir, 'tls.cert')
        self._macaroon_file = os.path.join(
            self.data_dir, 'data/chain/bitcoin/regtest/admin.macaroon')
        self._lnd_config_file = os.path.join(self.data_dir, 'lnd.conf')

        # lncli
        self.rpc_command = [
            self.rpc_binary,
            '--lnddir=' + self.data_dir,
            '--rpcserver=' + str(self._grpc_host),
            '--macaroonpath=' + self._macaroon_file,
            '--network=regtest',
        ]

    def start(self, from_scratch=True):
        """Start an LND node.

        :param from_scratch: bool
        :return:
        """

        if from_scratch:
            self.clear_directory()
            self.setup_datadir()
        else:
            if not os.path.isdir(self.data_dir):
                raise FileNotFoundError(
                    '{}: Lnd data directory not found '
                    '(from_scratch = False).'.format(self.name))

        command = [self.server_binary,
                   '--trickledelay=1',
                   '--lnddir=' + self.data_dir,
                   '--noseedbackup',
                   '--externalip=localhost:' + str(self.lnport)]

        cmd = ' '.join(command)
        logger.info("%s: Starting lnd: %s ", self.name, cmd)

        self.thread = threading.Thread(target=self.tail)
        self.thread.daemon = False

        # we start nonblocking with Popen
        self.server_process = subprocess.Popen(command, stdout=subprocess.PIPE)
        self.thread.start()
        self.running = True

        # we consider lnd to be started, when it has scanned the chain
        self.wait_for_log("Finished rescan")

        return self.server_process

    def stop(self):
        self.rpc(['stop'])
        logger.info('%s: stopped lnd', self.name)
        # should wait for the process to really stop
        self.server_process.communicate()
        # TODO: if this fails, force stop

    def print_rpc_command(self):
        """Prints the lncli command to use in the shell for testing."""
        cmd = ' '.join(self.rpc_command)
        logger.info("%s:", self.name)
        logger.info(cmd)

    def setup_datadir(self):
        """Sets up the lnd data folder."""
        pathlib.Path(self.data_dir).mkdir(parents=True, exist_ok=True)
        config = lnd_config_template.format(
            name=self.name,
            lnd_port=self.lnport,
            rest_port=self._rest_port,
            rpc_port=self._grpc_port,
            base_fee_msat=self.base_fee_msat,
            fee_rate=int(1E6 * self.fee_rate),
        )

        with open(self._lnd_config_file, 'w') as f:
            f.write(config)

    def clear_directory(self):
        """Deletes the lnd data directory of this node."""
        logger.debug("%s: Cleaning up lnd data directory.", self.name)
        try:
            shutil.rmtree(self.data_dir)
        except FileNotFoundError:
            logger.debug("%s: Directory already clean.", self.name)

    def getinfo(self):
        returncode, info = self.rpc(['getinfo'])
        return info

    def getaddress(self):
        returncode, address = self.rpc(['newaddress', 'p2wkh'])
        return address['address']

    def _connect(self, pubkey, host):
        logger.info("%s: Connecting to %s", self.name, pubkey)
        address = pubkey + '@' + host
        returncode, info = self.rpc(['connect', address])
        return info

    # TODO: somehow unify sync and async apis
    def _openchannel(self, pubkey, local_sat, remote_sat):
        logger.info("%s: Open channel to %s", self.name, pubkey)
        command = ['openchannel', '--min_confs', '0', pubkey, local_sat,
                   remote_sat]
        returncode, info = self.rpc(command)
        return info

    async def _a_openchannel(self, pubkey, local_sat, remote_sat):
        logger.info("%s: Async open channel to %s", self.name, pubkey)
        command = ['openchannel', '--min_confs', '0', pubkey, local_sat,
                   remote_sat]
        returncode, info = await self.async_grpc(command)
        return info

    def connect_and_openchannel(self, pubkey, host, local_sat, remote_sat):
        self._connect(pubkey, host)
        info = self._openchannel(pubkey, local_sat, remote_sat)
        funding_txid = info['funding_txid']
        return funding_txid

    def disconnect(self, pubkey):
        logger.info("%s: Disconnecting %s.", self.name, pubkey)
        command = ['disconnect', pubkey]
        returncode, info = self.rpc(command)
        return info

    def set_node_pubkey(self):
        info = self.getinfo()
        logger.info(
            "%s: setting node public key to %s",
            self.name, info['identity_pubkey'])
        self.pubkey = info['identity_pubkey']

    def listchannels(self) -> List[ChannelState]:
        command = ['listchannels']
        _, channels = self.rpc(command)
        channel_states = []
        for c in channels['channels']:
            funding_txid, outpoint = c['channel_point'].split(':')
            channel_states.append(ChannelState(
                capacity=c['capacity'],
                channel_id=c['chan_id'],
                commit_fee=c['commit_fee'],
                funding_txid=funding_txid,
                initiator=c['initiator'],
                local_balance=c['local_balance'],
                outpoint=int(outpoint),
                remote_balance=c['remote_balance'],
                remote_pubkey=c['remote_pubkey'],
                state='OPEN' if c['active'] else 'OPENING',
            ))
        return channel_states

    def updatechanpolicy(self, base_fee_msat, fee_rate, time_lock_delta=20,
                         channel_point=None):
        command = [
            'updatechanpolicy', int(base_fee_msat), fee_rate, time_lock_delta]
        if channel_point:
            command += channel_point
        returncode, info = self.rpc(command)
        return info

    def getnetworkinfo(self):
        command = ['getnetworkinfo']
        _, networkinfo = self.rpc(command)
        return networkinfo

    def describegraph(self) -> List[ChannelInfo]:
        command = ['describegraph']
        _, networkinfo = self.rpc(command)
        channels = []
        for c in networkinfo['edges']:
            # construct channel info
            channel_info = ChannelInfo(
                node1_key=c['node1_pub'],
                node2_key=c['node2_pub'],
                channel_id=int(c['channel_id'])
            )
            channels.append(channel_info)
        return channels

    def walletbalance(self):
        command = ['walletbalance']
        _, walletbalance = self.rpc(command)
        return walletbalance


class Electrum(LightningDaemon):
    def __init__(self, name: str, node_properties: dict, nodes_folder: str,
                 binary_folder: Optional[str] = None):
        super().__init__(name, node_properties, nodes_folder, binary_folder)
        self.name = name

        # network definitions
        self.lnport = node_properties['port']

        # fees
        self.base_fee_msat = node_properties['base_fee_msat']
        self.fee_rate = node_properties['fee_rate']

        # take binaries from path, if no binary folder is given
        if binary_folder is None:
            binary_folder = ''
        self.server_binary = os.path.join(binary_folder, 'electrum')
        self.client_binary = os.path.join(binary_folder, 'electrum')

        # check if executables can be found
        if shutil.which(self.server_binary) is None:
            raise FileNotFoundError(
                f"electrum executable not found: {self.server_binary}")

        # rpc command
        self.rpc_command = [
            self.client_binary,
            '-v',
            '-D' + self.data_dir,
            '--regtest',
        ]

    def start(self, from_scratch=True):
        """
        Start an electrum node.

        :param from_scratch: bool
        :return:
        """

        if from_scratch:
            self.clear_directory()
            self.setup_datadir()
        else:
            if not os.path.isdir(self.data_dir):
                raise FileNotFoundError(
                    '{}: Electrum data directory not found '
                    '(from_scratch = False).'.format(self.name))

        logger.info("%s: Configuring electrum.", self.name)

        self.rpc(['--offline', 'create'])
        self.rpc(['--offline', 'setconfig', 'log_to_file', 'true'])
        self.rpc(['--offline', 'setconfig', 'use_gossip', 'true'])
        self.rpc(['--offline', 'setconfig', 'lightning_listen',
                  'localhost:' + str(self.lnport)])
        self.rpc(['--offline', 'setconfig', 'oneserver', 'true'])
        self.rpc(['--offline', 'setconfig', 'server', '127.0.0.1:51001:t'])

        logger.info("%s: Starting electrum.", self.name)
        self.thread = threading.Thread(target=self.tail)
        self.thread.daemon = False

        # we start nonblocking with Popen
        daemon_command = [
            self.server_binary,
            '-v',
            '-D' + self.data_dir,
            '--regtest',
            'daemon',
        ]
        self.server_process = subprocess.Popen(
            daemon_command, stdout=subprocess.PIPE, stderr=subprocess.STDOUT)
        self.thread.start()
        self.running = True

        # we consider lnd to be started, when it has scanned the chain
        self.wait_for_log("connection established")
        self.rpc(['load_wallet'])

        return self.server_process

    def stop(self):
        self.rpc(['stop'])
        logger.info('%s: stopped electrum', self.name)
        # should wait for the process to really stop
        self.server_process.communicate()
        # TODO: if this fails, force stop

    def print_rpc_command(self):
        """
        Prints the rpc command to use in the shell for testing.
        """
        cmd = ' '.join(self.rpc_command)
        logger.info("%s:", self.name)
        logger.info(cmd)

    def setup_datadir(self):
        """
        Sets up the lnd data folder.
        """
        pathlib.Path(self.data_dir).mkdir(parents=True, exist_ok=True)

    def clear_directory(self):
        """
        Deletes the electrum data directory of this node.
    """
        logger.debug("%s: Cleaning up lnd data directory.", self.name)
        try:
            shutil.rmtree(self.data_dir)
        except FileNotFoundError:
            logger.debug("%s: Directory already clean.", self.name)

    def getinfo(self):
        returncode, info = self.rpc(['getinfo'])
        return info

    def getaddress(self):
        returncode, address = self.rpc(['getunusedaddress'])
        return address

    def connect_and_openchannel(self, pubkey, host, local_sat, remote_sat) \
            -> str:
        self._wait_for_funds()
        logger.info("%s: Open channel to %s@%s", self.name, pubkey, host)
        command = ['open_channel', '--push_amount', remote_sat / 1E8,
                   f"{pubkey}@{host}", local_sat / 1E8]
        returncode, info = self.rpc(command)
        funding_txid = info.split(':')[0]
        return funding_txid

    def _wait_for_funds(self):
        logger.info(f"{self.name}: waiting for funds")
        while True:
            time.sleep(0.5)
            _, balance = self.rpc(['getbalance'])
            if (float(balance['confirmed']) > 0 and
                    balance.get('unconfirmed', None) is None):
                break

    def wait_all_channels_open(self):
        while True:
            logger.info(f"{self.name}: waiting for channels to be open")
            time.sleep(1.0)
            channel_states = self.listchannels()
            try:
                for c in channel_states:
                    # if not all channels are open, wait
                    if c.state != 'OPEN':
                        raise Exception
            except Exception as e:
                logger.exception(e)
                continue

            # all channels are open
            logger.info(f"{self.name}: all channels open")
            return

    @property
    def nodeid(self) -> str:
        proc, connection_string = self.rpc(['nodeid'])
        nodeid = connection_string.split('@')[0]
        return nodeid

    def set_node_pubkey(self):
        node_pubkey = self.nodeid
        logger.info(
            "%s: setting node public key to %s",
            self.name, node_pubkey)
        self.pubkey = node_pubkey

    def listchannels(self) -> List[ChannelState]:
        command = ['list_channels']
        _, channels = self.rpc(command)
        channel_states = []
        for c in channels:
            # convert the electrum notation of the short channel id to
            # the integer representation
            try:
                block, trans, out = map(int, c['short_channel_id'].split('x'))
                chan_id = convert_short_channel_id_to_channel_id(
                    block, trans, out)
            except AttributeError:
                chan_id = None

            funding_txid, outpoint = c['channel_point'].split(':')
            channel_states.append(ChannelState(
                capacity=c['local_balance'] + c['remote_balance'],
                channel_id=chan_id,
                commit_fee=0,  # TODO: expose commit fee in electrum
                funding_txid=funding_txid,
                initiator=True,  # TODO: expose initiator in electrum
                local_balance=c['local_balance'],
                outpoint=int(outpoint),
                state=c['state'],
                remote_balance=c['remote_balance'],
                remote_pubkey=c['remote_pubkey']
            ))
        return channel_states

    def updatechanpolicy(self, base_fee_msat, fee_rate, time_lock_delta=20,
                         channel_point=None):
        command = [
            'updatechanpolicy', int(base_fee_msat), fee_rate, time_lock_delta]
        if channel_point:
            command += channel_point
        returncode, info = self.rpc(command)
        return info

    def getnetworkinfo(self):
        command = ['getnetworkinfo']
        _, networkinfo = self.rpc(command)
        return networkinfo

    def describegraph(self) -> List[ChannelInfo]:
        command = ['dumpgraph']
        _, networkinfo = self.rpc(command)
        channels = []
        for c in networkinfo['channels']:
            # convert short channel id to channel id
            cid: str = c['short_channel_id']
            cid: bytes = bfh(cid)
            blockheight = int.from_bytes(cid[:3], byteorder='big')
            transaction = int.from_bytes(cid[3:6], byteorder='big')
            output = int.from_bytes(cid[6:8], byteorder='big')
            channel_id = convert_short_channel_id_to_channel_id(
                blockheight, transaction, output)

            # construct channel info
            channel_info = ChannelInfo(
                node1_key=c['node1_id'],
                node2_key=c['node2_id'],
                channel_id=channel_id
            )
            channels.append(channel_info)
        return channels

    def walletbalance(self):
        command = ['walletbalance']
        _, walletbalance = self.rpc(command)
        return walletbalance



"""Contains Lightning network specific conversion utilities."""

import re
import time


def convert_short_channel_id_to_channel_id(blockheight, transaction, output):
    """
    Converts short channel id (blockheight:transaction:output) to a long integer channel id.

    :param blockheight:
    :param transaction: Number of transaction in the block.
    :param output: Number of output in the transaction.
    :return: channel id: Encoded integer number representing the channel,
     can be decoded by :func:`lib.conversion.extract_short_channel_id_from_string`.
    """
    return blockheight << 40 | transaction << 16 | output


def convert_channel_id_to_short_channel_id(channel_id):
    """
    Converts a channel id to blockheight, transaction, output
    """
    return channel_id >> 40, channel_id >> 16 & 0xFFFFFF, channel_id & 0xFFFF


def extract_short_channel_id_from_string(string):
    """
    Parses a payment error message for the short channel id of the form XXXX:XXXX:XXX.

    :param string:
    :return: short channel id [blockheight, transaction, output]
    """
    match = re.search(r'[0-9]+:[0-9]+:[0-9]+', string)
    group = match.group()
    groups = list(map(int, group.split(':')))
    assert len(groups) == 3
    return groups


def channel_unbalancedness_and_commit_fee(local_balance, capacity, commit_fee,
                                          initiator):
    """
    Calculates the unbalancedness.

    :param local_balance: int
    :param capacity: int
    :param commit_fee: int
    :param initiator: bool
    :return: float:
        in [-1.0, 1.0]
    """
    # inverse of the formula:
    # c.local_balance = c.capacity * 0.5 * (-unbalancedness + 1) - commit_fee
    commit_fee = 0 if not initiator else commit_fee
    return -(2 * float(local_balance + commit_fee) / capacity - 1), commit_fee


def height_to_timestamp(node, close_height):
    now = time.time()
    blocks_ago = node.blockheight - close_height
    time_ago = blocks_ago * 10 * 60
    timestamp_sec = int(now - time_ago)
    return timestamp_sec


from unittest import TestCase
from lndmanage.lib.ln_utilities import channel_unbalancedness_and_commit_fee


class LnUtilityTest(TestCase):
    def test_unbalancedness_formula(self):
        self.assertAlmostEqual(
            channel_unbalancedness_and_commit_fee(
                500000, 1000000, 1000, False)[0], 0.0)
        self.assertAlmostEqual(
            channel_unbalancedness_and_commit_fee(
                500000, 1000000, 1000, True)[0], -0.002)
        self.assertAlmostEqual(
            channel_unbalancedness_and_commit_fee(
                600000, 1000000, 0, False)[0], -0.2)



from argparse import ArgumentParser, ArgumentDefaultsHelpFormatter
import logging.config

from lnregtest.lib.network import Network
from lnregtest.lib.common import logger_config
from lnregtest import __version__

logging.config.dictConfig(logger_config)


def main():
    opts = parse_args()

    testnet = Network(
        from_scratch=opts.from_scratch,
        node_limit=opts.node_limit,
        network_definition_location=opts.network_definition,
        nodedata_folder=opts.nodedata_folder
    )
    testnet.run_continuously()


def parse_args():
    parser = ArgumentParser(
        description='lnregtest',
        formatter_class=ArgumentDefaultsHelpFormatter)
    parser.add_argument(
        '--from_scratch',
        action='store_true',
        help='If True, the network is generated from scratch.',
        default=True)
    parser.add_argument(
        '--node_limit',
        type=str,
        default='Z',
        help="Limits the number of nodes taken from the network definition, "
             "e.g. if you supply 'C', then nodes A, B, and C are created.")
    parser.add_argument(
        '--network_definition',
        type=str,
        default='star_ring',
        help='Defines the network (either absolute path or module name), '
             'examples can be found in lnregtest/network_definitions')
    parser.add_argument(
        '--nodedata_folder',
        type=str,
        default='',
        help='If nothing is specified, a temporary folder is created to store '
             'the runtime data. Otherwise, if an absolute path is given, '
             'the runtime data is stored there and can be restarted.')
    parser.add_argument(
        '--version',
        '-v',
        action='version',
        version="%(prog)s " + __version__)

    return parser.parse_args()



import json
import difflib


def format_dict(dictionary):
    """
    Formats dicts with indentation.

    :param dictionary: dict
    :return: str
    """
    return json.dumps(dictionary, indent=4)


def decode_byte_string_to_dict_or_str(out):
    """
    Takes output from Process and converts it to a dict.
    :param out: str
    :return: dict
    """
    try:
        json_data = json.loads(out)
        return json_data
    except json.decoder.JSONDecodeError:
        if type(out) == str:
            return out.strip()
        if type(out) == bytes:
            return out.decode().strip()


def dict_comparison(dict1, dict2, show_diff=False):
    """
    Compares two dicts for equality by converting to key-sorted dicts.

    The difference can be plotted by giving an additional show_diff bool.
    :param dict1: dict
    :param dict2: dict
    :param show_diff: bool
    :return: bool
    """
    dict1 = json.dumps(dict1, sort_keys=True, indent=4)
    dict2 = json.dumps(dict2, sort_keys=True, indent=4)
    are_equal = dict1 == dict2

    # only if the two dicts are not the same, show difference
    if not are_equal and show_diff:
        d = difflib.Differ()
        # need to split into lines by newline character but keep newline
        # for better printing
        dict1_lines = [l + '\n' for l in dict1.split('\n')]
        dict2_lines = [l + '\n' for l in dict2.split('\n')]
        difference = d.compare(dict1_lines, dict2_lines)
        print('\nDicts are NOT equal:')
        print(''.join(difference))

    return are_equal


def convert_short_channel_id_to_channel_id(blockheight, transaction, output) -> int:
    """
    Converts short channel id (blockheight:transaction:output) to a long integer channel id.

    :param blockheight:
    :param transaction: Number of transaction in the block.
    :param output: Number of output in the transaction.
    :return: channel id: Encoded integer number representing the channel,
     can be decoded by :func:`lib.conversion.extract_short_channel_id_from_string`.
    """
    return blockheight << 40 | transaction << 16 | output

# convert to bytes from hex
bfh = bytes.fromhex



import os

# define waiting periods
WAIT_AFTER_MINING_THREE = 0.5
WAIT_AFTER_ALL_LND_STARTED = 1
WAIT_AFTER_FILLING_WALLETS = 3
WAIT_BEFORE_CLEANUP = 1
WAIT_SYNC_BITCOIND = 1
WAIT_SYNC_ELECTRUMX = 1

common_path = os.path.dirname(os.path.realpath(__file__))
root_path = os.path.join(common_path, '../../')

logger_config = {
    'version': 1,
    'disable_existing_loggers': False,
    'formatters': {
        'file': {
            'format': '[%(asctime)s %(levelname)s] %(message)s',
            'datefmt': '%Y-%m-%d %H:%M:%S'
        },
        'standard': {
            'format': '%(message)s',
        },
    },
    'handlers': {
        'default': {
            'level': 'INFO',
            'formatter': 'standard',
            'class': 'logging.StreamHandler',
            'stream': 'ext://sys.stdout',  # Default is stderr
        },
        'file': {
            'level': 'DEBUG',
            'formatter': 'file',
            'class': 'logging.FileHandler',
            'filename': os.path.join(root_path, 'regtestnet.log'),
        },
    },
    'loggers': {
        '': {  # root logger
            'handlers': ['default', 'file'],
            'level': 'DEBUG',
            'propagate': True
        },
    }
}



import re, setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

with open('./lndmanage/__init__.py', 'r') as f:
    MATCH_EXPR = "__version__[^'\"]+(['\"])([^'\"]+)"
    VERSION = re.search(MATCH_EXPR, f.read()).group(2)

# package:
# (venv) pip install pep517 setuptools wheel sdist twine
# (venv) python3 -m pep517.build --source --binary .
# upload:
# (venv) twine upload --repository testpypi dist/*

setuptools.setup(
    name="lndmanage",
    version=VERSION,
    author="bitromortac",
    author_email="bitromortac@protonmail.com",
    description="Channel management tool for lightning network daemon (LND) "
                "operators.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/bitromortac/lndmanage",
    packages=setuptools.find_packages(),
    install_requires=[
        'cycler==0.10.0',
        'decorator==4.4.2',
        'googleapis-common-protos==1.52.0',
        'grpcio==1.31.0',
        'grpcio-tools==1.31.0',
        'kiwisolver==1.2.0',
        'networkx==2.4',
        'numpy==1.19.1',
        'protobuf==3.12.4',
        'Pygments==2.7.4',
        'pyparsing==2.4.7',
        'python-dateutil==2.8.1',
        'six==1.15.0',
    ],
    include_package_data=True,
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    entry_points={
        "console_scripts": [
            "lndmanage = lndmanage.lndmanage:main",
        ]
    },
)
